-------------------------------------------------------------------------------
-- Locals                                                                    --
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.0"):new("ClosetGnome")
local gratuity = AceLibrary("Gratuity-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local tablet = AceLibrary("Tablet-2.0")

local queuedSet = nil
local lastEquippedSet = nil

local _G = getfenv(0)

local slots = {
	{slot = "Head"},
	{slot = "Shoulder"},
	{slot = "Chest"},
	{slot = "Waist"},
	{slot = "Legs"},
	{slot = "Feet"},
	{slot = "Wrist"},
	{slot = "Hands"},
	{slot = "MainHand"},
	{slot = "SecondaryHand"},
	{slot = "Ranged"},
	{slot = "Ammo"},
	{slot = "Neck"},
	{slot = "Back"},
	{slot = "Finger0"},
	{slot = "Finger1"},
	{slot = "Trinket0"},
	{slot = "Trinket1"},
	{slot = "Relic"},
}

-------------------------------------------------------------------------------
-- Localization                                                              --
-------------------------------------------------------------------------------

L:RegisterTranslations("enUS", function() return {
	["ClosetGnome"] = true,
	["ClosetGnome options."] = true,
	["Enable or disable the slots you want in this set by clicking the items in your character frame. Green slots are enabled, red are disabled. Type in the set name below, and click Add."] = true,
	["Add"] = true,
	["Creates a new set."] = true,
	["Wear"] = true,
	["Change equipment set."] = true,
	["Cancel"] = true,
	["Always equip weapons"] = true,
	["Equip weapons in sets even if you are in combat."] = true,
	["Equip %s."] = true,
	["Delete"] = true,
	["Enter the set name to delete it."] = true,
	["<set name>"] = true,
	["Couldn't find %s in your inventory."] = true,

	["In combat: %s queued for swap."] = true,
	["Added set: %s."] = true,
	["Deleted set: %s."] = true,
	["Updating set: %s."] = true,

	["Last equipped:"] = true,
	["Available sets"] = true,
	["Click a set above to change your equipment."] = true,
	["You have not created any equipment sets yet, please right-click the icon and click 'Add' to create a set."] = true,

	-- Random quips when you equip a set.
	[1] = "Your ClosetGnome rapidly dresses you up in %s before he runs off with another sock.",
	[2] = "Once in a blue moon your ClosetGnome will equip you with an Artifact item. Not today, however, you'll have to settle for %s.",
	[3] = "ClosetGnomes are tidy creatures. Not yours, though, as he equips %s he just piles your old outfit in the corner.",
	[4] = "It's almost as if he dances around you while he dresses you up in %s. Although it's a very ugly dance.",
	[5] = "Since you forgot to feed your ClosetGnome today, he expresses his anger by equipping your %s backwards.",
	[6] = "Never let your ClosetGnome tend to other needs than %s.",
	[7] = "%s looks good on you today, at least compared to how it looked on your ClosetGnome when you caught him wearing it.",
	[8] = "ClosetGnome equips %s.",
	[9] = "Your ClosetGnome flinches in disgust as he equips %s. You're not sure why, but what's that smell ...",
	[10] = "Is that a whirlwind?! No, it's ClosetGnome dressing you in %s!",
	[11] = "Recently returned from the ClosetGnome convention, he's got a lot to talk about while dressing you in %s.",
	[12] = "Noone knows if ClosetGnomes party, but you do. This is how it looked last time you dressed yourself in %s after a party.",
	[13] = "Getting' Jiggy With It was never your favorite, but you have to admit that it fits nicely with %s.",
	[14] = "Even though your washer is clearly labeled 'Certified by ClosetGnomes, Inc.', it's clear that %s has not been cleaned in a while.",
	[15] = "%s has the same color socks as your last set, and the one before that, and before that.",
	[16] = "Instead of equipping %s, your ClosetGnome starts gnawing on your leg. You shake him off and do it yourself.",
	[17] = "Noone has ever suspected you of having a ClosetGnome. Most of them are actually good at equipping things like %s.",
} end)

-------------------------------------------------------------------------------
-- Addon declaration                                                         --
-------------------------------------------------------------------------------

ClosetGnome = AceLibrary("AceAddon-2.0"):new("AceHook-2.0", "AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0", "FuBarPlugin-2.0")
ClosetGnome:RegisterDB("ClosetGnomeDB", "ClosetGnomePerCharDB")
ClosetGnome:RegisterDefaults("char", {
	set = {},
	alwaysWeapons = true,
})
ClosetGnome.OnMenuRequest = {
	type = "group",
	name = L["ClosetGnome"],
	desc = L["ClosetGnome options."],
	args = {
		add = {
			type = "execute",
			name = L["Add"],
			desc = L["Creates a new set."],
			func = function() ClosetGnome:CreateSet() end,
			order = 101,
		},
		wear = {
			type = "group",
			name = L["Wear"],
			desc = L["Change equipment set."],
			args = {},
			order = 102,
			disabled = function() return next(ClosetGnome.OnMenuRequest.args.wear.args) == nil end
		},
		spacer = { type = "header", order = 103 },
		alwaysweapons = {
			type = "toggle",
			name = L["Always equip weapons"],
			desc = L["Equip weapons in sets even if you are in combat."],
			get = function() return ClosetGnome.db.char.alwaysWeapons end,
			set = function(v) ClosetGnome.db.char.alwaysWeapons = v end,
			order = 104,
		},
		delete = {
			type = "text",
			name = L["Delete"],
			desc = L["Enter the set name to delete it."],
			usage = L["<set name>"],
			get = false,
			set = function(v) ClosetGnome:DeleteSet(v) end,
			order = 105,
			disabled = function() return next(ClosetGnome.OnMenuRequest.args.wear.args) == nil end
		},
	},
}
ClosetGnome:RegisterChatCommand({"/closetgnome", "/cg"}, ClosetGnome.OnMenuRequest)

-- FuBar stuff.
--ClosetGnome.name = L["ClosetGnome"]
ClosetGnome.hideWithoutStandby = true
ClosetGnome.hasIcon = "Interface\\Icons\\INV_Chest_Cloth_17"
ClosetGnome.defaultMinimapPosition = 240
ClosetGnome.clickableTooltip  = true

-------------------------------------------------------------------------------
-- Initialization                                                            --
-------------------------------------------------------------------------------

function ClosetGnome:OnEnable()
	StaticPopupDialogs["ClosetGnomeAdd"] = {
		text = L["Enable or disable the slots you want in this set by clicking the items in your character frame. Green slots are enabled, red are disabled. Type in the set name below, and click Add."],
		button1 = L["Add"],
		button2 = L["Cancel"],
		OnCancel = function() self:CancelCreateSet() end,
		sound = "levelup2",
		whileDead = 1,
		hideOnEscape = 1,
		timeout = 0,
		OnShow = function()
			-- We have to do this onshow to reset the previous text
			getglobal(this:GetName().."EditBox"):SetText("")
		end,
		OnAccept = function()
			local name = getglobal(this:GetParent():GetName().."EditBox"):GetText()
			name = string.gsub(name, " ", "")
			self:AddSetFromDoll(name)
		end,
		hasEditBox = 1,
	}

	-- Temporary table for which slots to use when adding a new set.
	self.activeSlots = {}

	-- Create the paper doll slot overlays, most of the code credit to Fizzle.
	for _, item in ipairs(slots) do
		local gslot = _G["Character" .. item.slot .. "Slot"]
		if gslot and not item.frame then
			local border = gslot:CreateTexture(gslot:GetName() .. "ClosetGnome", "OVERLAY")
			border:SetTexture("Interface\\Buttons\\UI-ActionButton-Border")
			border:SetVertexColor(0, 255, 0)
			border:SetBlendMode("ADD")
			border:SetAlpha(0.75)
			border:SetHeight(68)
			border:SetWidth(68)
			border:SetPoint("CENTER", gslot, "CENTER", 0, 1)
			border:Hide()
			item.frame = border
		end
	end

	self:RegisterEvent("AceEvent_FullyInitialized")

	self:UpdateSetMenu()
	self:UpdateTooltip()
end

-------------------------------------------------------------------------------
-- Menu                                                                      --
-------------------------------------------------------------------------------

function ClosetGnome:UpdateSetMenu()
	for i in pairs(self.OnMenuRequest.args.wear.args) do
		self.OnMenuRequest.args.wear.args[i] = nil
	end
	if table.setn then table.setn(self.OnMenuRequest.args.wear.args, 0) end

	self.OnMenuRequest.args.wear.args = {}
	for k,v in pairs(self.db.char.set) do
		local set = k
		self.OnMenuRequest.args.wear.args[k] = {}
		self.OnMenuRequest.args.wear.args[k].type = "execute"
		self.OnMenuRequest.args.wear.args[k].name = set
		self.OnMenuRequest.args.wear.args[k].desc = string.format(L["Equip %s."], set)
		self.OnMenuRequest.args.wear.args[k].func = function() ClosetGnome:WearSet(set) end
	end
end

-------------------------------------------------------------------------------
-- Addon methods                                                             --
-------------------------------------------------------------------------------

-- Opens the character frame and lets the user uncheck the items he doesn't
-- want.
function ClosetGnome:CreateSet()
	self.activeSlots = {}
	for _, item in ipairs(slots) do
		if item.frame then
			item.frame:SetVertexColor(0, 255, 0)
			item.frame:Show()
			local gslot = _G["Character" .. item.slot .. "Slot"]
			self.activeSlots[gslot:GetName()] = true
			item.script = gslot:GetScript("OnClick")
			gslot:SetScript("OnClick", function()
				ClosetGnome.activeSlots[this:GetName()] = not ClosetGnome.activeSlots[this:GetName()]
				if ClosetGnome.activeSlots[this:GetName()] then
					getglobal(this:GetName().."ClosetGnome"):SetVertexColor(0, 255, 0)
				else
					getglobal(this:GetName().."ClosetGnome"):SetVertexColor(255, 0, 0)
				end
			end)
		end
	end
	ShowUIPanel(CharacterFrame)
	CharacterFrame_ShowSubFrame("PaperDollFrame")
	OpenAllBags()
	self:Hook("CharacterFrame_OnHide")
	self:Hook("CharacterFrameTab_OnClick")

	dewdrop:Close()
	StaticPopup_Show("ClosetGnomeAdd")
end

function ClosetGnome:CancelCreateSet()
	HideUIPanel(CharacterFrame)
	if self:IsHooked("CharacterFrame_OnHide") then
		self:Unhook("CharacterFrame_OnHide")
	end
end

function ClosetGnome:AddSetFromDoll(name)
	local slots = {}
	for slot, active in pairs(self.activeSlots) do
		if active then
			table.insert(slots, string.sub(slot, 10))
		end
	end
	self:AddSet(name, slots)
	HideUIPanel(CharacterFrame)
	self:CharacterFrame_OnHide()
end

function ClosetGnome:AddSet(name, activeSlots)
	local text = string.format(L["Added set: %s."], name)
	local event = "ClosetGnome_AddSet"
	if self:HasSet(name) then
		text = string.format(L["Updating set: %s."], name)
		event = "ClosetGnome_UpdateSet"
	end

	for i, slot in pairs(activeSlots) do
		local id = GetInventorySlotInfo(slot)
		local hasItem = gratuity:SetInventoryItem("player", id)
		if not self.db.char.set[name] then self.db.char.set[name] = {} end
		if hasItem then
			local itemName = gratuity:GetLine(1)
			self.db.char.set[name][id] = itemName
		else
			self.db.char.set[name][id] = false
		end
	end

	lastEquippedSet = name
	self:TriggerEvent(event, name)
	self:Print(text)

	self:UpdateSetMenu()
	self:UpdateTooltip()
end

function ClosetGnome:DeleteSet(name)
	if self.db.char.set[name] then
		for i in pairs(self.db.char.set[name]) do
			self.db.char.set[name][i] = nil
		end
		set = nil
		self.db.char.set[name] = nil

		self:TriggerEvent("ClosetGnome_DeleteSet", name)
		self:Print(string.format(L["Deleted set: %s."], name))
		self:UpdateSetMenu()
		self:UpdateTooltip()
	end
end

function ClosetGnome:WearSet(name)
	if CursorHasItem() then return end

	local set = self.db.char.set[name]
	if not set then return end

	dewdrop:Close()

	if UnitAffectingCombat("player") and not self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		if self.db.char.alwaysWeapons then
			for slot, item in pairs(set) do
				if slot == 16 or slot == 17 or slot == 18 then
					gratuity:SetInventoryItem("player", slot)
					local currentItem = gratuity:GetLine(1)
					if item ~= currentItem then
						self:EquipItem(slot, item)
					end
				end
			end
			lastEquippedSet = name
			self:UpdateTooltip()
			if self:IsSetFullyEquipped(set) then
				self:TriggerEvent("ClosetGnome_WearSet", name)
				return
			else
				self:TriggerEvent("ClosetGnome_PartlyWearSet", name)
			end
		end
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:Print(string.format(L["In combat: %s queued for swap."], name))
		queuedSet = name
		return
	end
	if self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end

	queuedSet = nil
	local rand = math.random(1, 17)
	-- I don't want the addon name in front for this one.
	-- Thanks to MrPlow for the color :)
	if L[rand] ~= nil then
		DEFAULT_CHAT_FRAME:AddMessage("|cffeda55f"..string.format(L[rand], "|cffd9d919"..name.."|cffeda55f").."|r")
	end

	lastEquippedSet = name
	for slot, item in pairs(set) do
		gratuity:SetInventoryItem("player", slot)
		local currentItem = gratuity:GetLine(1)
		if item ~= currentItem then
			self:EquipItem(slot, item)
		end
	end
	self:TriggerEvent("ClosetGnome_WearSet", name)

	self:UpdateTooltip()
end

-------------------------------------------------------------------------------
-- Utility                                                                   --
-------------------------------------------------------------------------------

function ClosetGnome:IsSetFullyEquipped(set)
	for slot, item in pairs(set) do
		gratuity:SetInventoryItem("player", slot)
		local currentItem = gratuity:GetLine(1)
		if item ~= currentItem then
			return false
		end
	end
	return true
end

function ClosetGnome:EquipItem(slot, item, secondTry)
	if item == false then
		-- We need to find a free slot in the inventory for this item.
		-- If there's anything in the slot.
		local hasItem = gratuity:SetInventoryItem("player", slot)
		if hasItem then
			PickupInventoryItem(slot)
			for theBag = 0, NUM_BAG_FRAMES, 1 do
				local numSlot = GetContainerNumSlots(theBag)
				for theSlot = 1, numSlot, 1 do
					local texture, itemCount, locked = GetContainerItemInfo(theBag, theSlot)
					if not texture then
						PickupContainerItem(theBag, theSlot)
						return true
					end
				end
			end
			AutoEquipCursorItem()
			if not secondTry then
				self:ScheduleEvent("cg"..slot..item, self.EquipItem, 0.1, self, slot, item, true)
			end
			return false
		end
		return true
	else
		local bagNum, slotNum = self:FindItem(item)
		if bagNum > -1 then
			PickupContainerItem(bagNum, slotNum)
			if slot < 11 or slot == 15 or slot == 19 then
				AutoEquipCursorItem()
			else
				EquipCursorItem(slot)
			end
			return true
		else
			if not secondTry then
				self:ScheduleEvent("cg"..slot..item, self.EquipItem, 0.1, self, slot, item, true)
			else
				self:Print(string.format(L["Couldn't find %s in your inventory."], item))
			end
		end
	end
	return false
end

function ClosetGnome:HasSet(name)
	return self.db.char.set[name] ~= nil
end

function ClosetGnome:FindItem(item)
	for i = NUM_BAG_FRAMES, 0, -1 do
		for j = GetContainerNumSlots(i), 1, -1 do
			gratuity:SetBagItem(i, j)
			local bagItem = gratuity:GetLine(1)
			if bagItem == item then return i, j end
		end
	end
	return -1
end

-------------------------------------------------------------------------------
-- Hooks                                                                     --
-------------------------------------------------------------------------------

-- Only hooked if we are showing the paper doll frame with set adding
function ClosetGnome:CharacterFrame_OnHide()
	for _, item in ipairs(slots) do
		if item.frame then
			item.frame:Hide()
			local gslot = _G["Character" .. item.slot .. "Slot"]
			self.activeSlots[gslot:GetName()] = nil
			gslot:SetScript("OnClick", item.script)
		end
	end

	StaticPopup_Hide("ClosetGnomeAdd")
	if self:IsHooked("CharacterFrame_OnHide") then
		self:Unhook("CharacterFrame_OnHide")
	end
	if self:IsHooked("CharacterFrameTab_OnClick") then
		self:Unhook("CharacterFrameTab_OnClick")
	end
end

function ClosetGnome:CharacterFrameTab_OnClick() --[[ noop ]] end

-------------------------------------------------------------------------------
-- Events                                                                    --
-------------------------------------------------------------------------------

function ClosetGnome:AceEvent_FullyInitialized()
	-- Loop through the current sets and check if we have any of them equipped.
	for k,v in pairs(self.db.char.set) do
		if self:IsSetFullyEquipped(v) then
			lastEquippedSet = k
			self:SetText(k)
			self:TriggerEvent("ClosetGnome_WearSet", k)
			break
		end
	end
	self:UpdateText()
end

function ClosetGnome:PLAYER_REGEN_ENABLED()
	if not queuedSet then return end
	self:WearSet(queuedSet)
end

-------------------------------------------------------------------------------
-- FuBarPlugin                                                               --
-------------------------------------------------------------------------------

function ClosetGnome:OnTooltipUpdate()
	if lastEquippedSet then
		local current = tablet:AddCategory(
			"columns", 2,
			"showWithoutChildren", false,
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text2R", 1,
			"child_text2G", 1,
			"child_text2B", 1
		)
		current:AddLine("text", L["Last equipped:"], "text2", lastEquippedSet, "justify", "LEFT", "justify2", "RIGHT")
		self:SetText(lastEquippedSet)
	else
		self:SetText(L["ClosetGnome"])
	end

	local cat = nil
	for k,v in pairs(self.db.char.set) do
		if not cat then
			cat = tablet:AddCategory("columns", 1, "text", L["Available sets"], "showWithoutChildren", false)
		end
		local set = k
		cat:AddLine("text", set, "func", function(set) ClosetGnome:WearSet(set) end, "arg1", set, "justify", "LEFT")
	end

	if cat ~= nil then
		tablet:SetHint(L["Click a set above to change your equipment."])
	else
		tablet:SetHint(L["You have not created any equipment sets yet, please right-click the icon and click 'Add' to create a set."])
	end
end

